#include <windows.h>
#include <stdio.h>
#include "memory.h"
#include "bcCommon.h"
#include "bcsvc.h"

static CRITICAL_SECTION InfoSynchro;
static struct InfoRec *Info=NULL;
static DWORD InfoCount=0;
static DWORD MaxError=20;
static DWORD ThreadCount=4;
static HANDLE StartEvent=NULL;
static char user[256];
static char domain[256];
static char password[256];
static DWORD LogonType=LOGON32_LOGON_INTERACTIVE;

void InitInfo(void)
{
  HKEY hKey; DWORD Type,DataSize;
  if((RegOpenKeyEx(HKEY_LOCAL_MACHINE,"SYSTEM\\CurrentControlSet\\Services\\"SVC_NAME"\\Parameters",0,KEY_QUERY_VALUE,&hKey))==ERROR_SUCCESS)
  {
    DataSize=sizeof(MaxError);
    RegQueryValueEx(hKey,"MaxError",0,&Type,(LPBYTE)&MaxError,&DataSize);
    DataSize=sizeof(ThreadCount);
    RegQueryValueEx(hKey,"ThreadCount",0,&Type,(LPBYTE)&ThreadCount,&DataSize);
    if((ThreadCount>16)||(ThreadCount<1)) ThreadCount=4;
    memset(user,0,sizeof(user));
    memset(domain,0,sizeof(domain));
    memset(password,0,sizeof(password));
    DataSize=sizeof(user);
    RegQueryValueEx(hKey,"LogonUser",0,&Type,(LPBYTE)user,&DataSize);
    DataSize=sizeof(domain);
    RegQueryValueEx(hKey,"LogonDomain",0,&Type,(LPBYTE)domain,&DataSize);
    DataSize=sizeof(password);
    RegQueryValueEx(hKey,"LogonPassword",0,&Type,(LPBYTE)password,&DataSize);
    DataSize=sizeof(LogonType);
    RegQueryValueEx(hKey,"LogonType",0,&Type,(LPBYTE)&LogonType,&DataSize);
    RegCloseKey(hKey);
    if((LogonType!=LOGON32_LOGON_INTERACTIVE)&&(LogonType!=LOGON32_LOGON_BATCH)&&(LogonType!=LOGON32_LOGON_SERVICE))
      LogonType=LOGON32_LOGON_INTERACTIVE;
  }
  StartEvent=CreateEvent(NULL,TRUE,FALSE,NULL);
  InitializeCriticalSection(&InfoSynchro);
}

void FreeInfo(void)
{
  ResetStartEvent();
  SetAllStopInfo();
  WaitForInfo();
  DeleteCriticalSection(&InfoSynchro);
  if(Info) free(Info);
  Info=NULL; InfoCount=0;
}

void ResetStartEvent(void)
{
  if(StartEvent)
  {
    SetEvent(StartEvent);
    CloseHandle(StartEvent); 
    StartEvent=NULL; 
  }
}

void WaitStartEvent(void)
{
  if(StartEvent)
    while(WaitForSingleObject(StartEvent,3000)!=WAIT_OBJECT_0)
      ReportStatusToSCMgr(SERVICE_START_PENDING,NO_ERROR,3000);
}

void AddInfo(DWORD type,wchar_t *Src,BOOL InfoEx)
{
  DWORD ThreadId=GetCurrentThreadId();
  EnterCriticalSection(&InfoSynchro);
  Info=(struct InfoRec *)realloc(Info,sizeof(struct InfoRec)*(InfoCount+1));
  if(Info)
  {
    Info[InfoCount].info.ThreadId=ThreadId;
    Info[InfoCount].info.type=type;
    wcscpy(Info[InfoCount].info.Src,Src);
    Info[InfoCount].info.InfoEx=InfoEx;
    wcscpy(Info[InfoCount].Src,L"");
    wcscpy(Info[InfoCount].Dest,L"");
    Info[InfoCount].stop=FALSE;
    Info[InfoCount].pause=FALSE;
    Info[InfoCount].TotalSize=0;
    Info[InfoCount].CurrentSize=0;
    Info[InfoCount].CurrentSizeAdd=0;
    Info[InfoCount].Errors=0;
    Info[InfoCount].PauseTime=0;

    InfoCount++;
  }
  else
    InfoCount=0;
  LeaveCriticalSection(&InfoSynchro);
}

void DelInfo(void)
{
  int i; DWORD ThreadId=GetCurrentThreadId();
  EnterCriticalSection(&InfoSynchro);
  for(i=0;i<InfoCount;i++)
  {
    if(Info[i].info.ThreadId==ThreadId)
    {
      memmove(Info+i,Info+i+1,sizeof(struct InfoRec)*(InfoCount-i-1));
      Info=(struct InfoRec *)realloc(Info,sizeof(struct InfoRec)*(InfoCount-1));
      if(Info)
        InfoCount--;
      else
        InfoCount=0;
      break;
    }
  }
  LeaveCriticalSection(&InfoSynchro);
}

DWORD GetAllInfo(struct SmallInfoRec **GetInfo)
{
  int i; DWORD Result=0;
  EnterCriticalSection(&InfoSynchro);
  if(InfoCount)
  {
    *GetInfo=(struct SmallInfoRec *)malloc(sizeof(struct SmallInfoRec)*InfoCount);
    if(*GetInfo)
    {
      Result=InfoCount;
      for(i=0;i<InfoCount;i++)
        (*GetInfo)[i]=Info[i].info;
    }
  }
  LeaveCriticalSection(&InfoSynchro);
  return Result;
}

void GetInfo(DWORD ThreadId,struct InfoRec *GetInfo)
{
  int i;
  EnterCriticalSection(&InfoSynchro);
  for(i=0;i<InfoCount;i++)
  {
    if(Info[i].info.ThreadId==ThreadId)
    {
      *GetInfo=Info[i];
      break;
    }
  }
  LeaveCriticalSection(&InfoSynchro);
}

static void CheckPause(BOOL *Stop)
{
  if(!*Stop)
  {
    int i; DWORD ThreadId=GetCurrentThreadId(); BOOL wait=TRUE;
    while(wait)
    {
      wait=FALSE;
      EnterCriticalSection(&InfoSynchro);
      for(i=0;i<InfoCount;i++)
      {
        if(Info[i].info.ThreadId==ThreadId)
        {
          *Stop=Info[i].stop;
          wait=Info[i].pause&(!Info[i].stop);
          Info[i].PauseTime+=2000000ULL;
          break;
        }
      }
      LeaveCriticalSection(&InfoSynchro);
      Sleep(200);
    }
  }
}

BOOL UpdateInfo(wchar_t *newSrc,wchar_t *newDest)
{
  int i; DWORD ThreadId=GetCurrentThreadId(); BOOL Result=FALSE,Pause=FALSE;
  EnterCriticalSection(&InfoSynchro);
  for(i=0;i<InfoCount;i++)
  {
    if(Info[i].info.ThreadId==ThreadId)
    {
      Result=Info[i].stop;
      Pause=Info[i].pause;
      if(!Result)
      {
        wcscpy(Info[i].Src,newSrc);
        wcscpy(Info[i].Dest,newDest);
      }
      break;
    }
  }
  LeaveCriticalSection(&InfoSynchro);
  if(Pause) CheckPause(&Result);
  return Result;
}

void SetStopInfo(DWORD ThreadId)
{
  int i;
  EnterCriticalSection(&InfoSynchro);
  for(i=0;i<InfoCount;i++)
  {
    if(Info[i].info.ThreadId==ThreadId)
    {
      Info[i].stop=TRUE;
      break;
    }
  }
  LeaveCriticalSection(&InfoSynchro);
}

void SetPauseInfo(DWORD ThreadId)
{
  int i;
  EnterCriticalSection(&InfoSynchro);
  for(i=0;i<InfoCount;i++)
  {
    if(Info[i].info.ThreadId==ThreadId)
    {
      Info[i].pause=!Info[i].pause;
      break;
    }
  }
  LeaveCriticalSection(&InfoSynchro);
}

void SetStartInfo(unsigned long long size)
{
  int i; DWORD ThreadId=GetCurrentThreadId();
  EnterCriticalSection(&InfoSynchro);
  for(i=0;i<InfoCount;i++)
  {
    if(Info[i].info.ThreadId==ThreadId)
    {
      Info[i].TotalSize=size;
      GetLocalTime(&Info[i].StartTime);
      break;
    }
  }
  LeaveCriticalSection(&InfoSynchro);
}

BOOL UpdatePosInfo(unsigned long long add,unsigned long long subadd)
{
  int i; DWORD ThreadId=GetCurrentThreadId(); BOOL Result=FALSE,Pause=FALSE;
  EnterCriticalSection(&InfoSynchro);
  for(i=0;i<InfoCount;i++)
  {
    if(Info[i].info.ThreadId==ThreadId)
    {
      Info[i].CurrentSize+=add;
      Info[i].CurrentSizeAdd=subadd;
      Result=Info[i].stop;
      Pause=Info[i].pause;
      break;
    }
  }
  LeaveCriticalSection(&InfoSynchro);
  if(Pause) CheckPause(&Result);
  return Result;
}

DWORD CheckErrorInfo(void)
{
  int i; DWORD ThreadId=GetCurrentThreadId(); DWORD Result=0;
  EnterCriticalSection(&InfoSynchro);
  for(i=0;i<InfoCount;i++)
  {
    if(Info[i].info.ThreadId==ThreadId)
    {
      Info[i].Errors++;
      if(Info[i].Errors==MaxError) Result=1;
      if(Info[i].Errors>MaxError) Result=2;
      break;
    }
  }
  LeaveCriticalSection(&InfoSynchro);
  return Result;
}

void SetAllStopInfo(void)
{
  int i;
  EnterCriticalSection(&InfoSynchro);
  for(i=0;i<InfoCount;i++)
    Info[i].stop=TRUE;
  LeaveCriticalSection(&InfoSynchro);
}

void WaitForInfo(void)
{
  BOOL wait=TRUE;
  while(wait)
  {
    EnterCriticalSection(&InfoSynchro);
    wait=InfoCount;
    LeaveCriticalSection(&InfoSynchro);
    Sleep(200);
  }
}

DWORD GetMaxError(void)
{
  return MaxError;
}

DWORD GetThreadCount(void)
{
  return ThreadCount;
}

char *GetLogonUser(void)
{
  return user;
}

char *GetLogonDomain(void)
{
  return domain;
}

char *GetLogonPassword(void)
{
  return password;
}

DWORD GetLogonType(void)
{
  return LogonType;
}
