#include <windows.h>
#include <winioctl.h>
#include <stdio.h>
#include "memory.h"
#include "bcCommon.h"
#include "bcsvc.h"

enum {retError, retOk, retStop};

unsigned long long MyGetFileSize(wchar_t *file);

DWORD WINAPI CopyProgressRoutine(LARGE_INTEGER TotalFileSize,
                                 LARGE_INTEGER TotalBytesTransferred,
                                 LARGE_INTEGER StreamSize,
                                 LARGE_INTEGER StreamBytesTransferred,
                                 DWORD dwStreamNumber, DWORD dwCallbackReason,
                                 HANDLE hSourceFile, HANDLE hDestinationFile, LPVOID lpData)
{
  DWORD Result=PROGRESS_CONTINUE;
  if(dwStreamNumber==1)
    if(UpdatePosInfo(0,StreamBytesTransferred.QuadPart))
      Result=PROGRESS_CANCEL;
  return Result;
}
   
static void MyRemoveDirectoryW(wchar_t *src)
{
  SetFileAttributesW(src,0);
  if(!RemoveDirectoryW(src)) Log(L"RemoveDirectoryW",src,NULL);
}

static void MyCopySecurityW(wchar_t *src,wchar_t *dest,BOOL access,BOOL dir)
{
  if(dir)
  {
    WIN32_FILE_ATTRIBUTE_DATA attr; HANDLE dest_handle;
    if(GetFileAttributesExW(src,GetFileExInfoStandard,&attr))
    {
      if(!SetFileAttributesW(dest,attr.dwFileAttributes))
        Log(L"SetFileAttributesW",dest,NULL);
      dest_handle=CreateFileW(dest,GENERIC_WRITE,0,NULL,OPEN_EXISTING,FILE_FLAG_BACKUP_SEMANTICS,NULL);
      if(dest_handle!=INVALID_HANDLE_VALUE)
      {
        if(attr.dwFileAttributes&FILE_ATTRIBUTE_COMPRESSED)
        {
          unsigned short type=COMPRESSION_FORMAT_DEFAULT; DWORD ret;
          if(!DeviceIoControl(dest_handle,FSCTL_SET_COMPRESSION,&type,sizeof(type),NULL,0,&ret,NULL))
            Log(L"DeviceIoControl",dest,NULL);
        }
        if(!SetFileTime(dest_handle,&attr.ftCreationTime,&attr.ftLastAccessTime,&attr.ftLastWriteTime))
          Log(L"SetFileTime",dest,NULL);
        CloseHandle(dest_handle);
      }
    }
    else
      Log(L"GetFileAttributesExW",src,NULL);
  }
  if(access)
  {
    DWORD needed; void *SD;
    if(!GetFileSecurityW(src,DACL_SECURITY_INFORMATION,NULL,0,&needed))
      if(GetLastError()==ERROR_INSUFFICIENT_BUFFER)
      {
        SD=malloc(needed);
        if(SD)
          if(GetFileSecurityW(src,DACL_SECURITY_INFORMATION,SD,needed,&needed))
            SetFileSecurityW(dest,DACL_SECURITY_INFORMATION,SD);
      }
  }
}

static DWORD MyCopyFileW(wchar_t *src,wchar_t *dest,BOOL fail,BOOL access,BOOL move)
{
  DWORD ret=TRUE; BOOL copyCancel=FALSE; int type=0;
  wchar_t *errname=L"MoveFileExW";
  if(UpdateInfo(src,dest)) return retStop;
  if(!move) type=1;
  if(!(*dest)) type=2;
  switch(type)
  {
    case 0: //move
      ret=MoveFileExW(src,dest,fail?0:MOVEFILE_REPLACE_EXISTING);
      if(ret)
        break;
      if(GetLastError()!=ERROR_NOT_SAME_DEVICE)
        break;
    case 1: //copy
      errname=L"CopyFileExW";
      ret=CopyFileExW(src,dest,CopyProgressRoutine,NULL,&copyCancel,fail?COPY_FILE_FAIL_IF_EXISTS:0);
      if(ret) MyCopySecurityW(src,dest,access,FALSE);
      if((!ret)||type)
      {
        if((!ret)&&(GetLastError()==ERROR_REQUEST_ABORTED))
          ret=TRUE;
        break;
      }
    case 2: //delete
      errname=L"DeleteFileW";
      ret=DeleteFileW(src);
      break;
  }
  if(ret)
    return retOk;
  else
  {
    Log(errname,src,dest);
    return retError;
  }
}

#define MY_DELETE ((flags&COPYFLAG_DELETE)!=0)

#define MY_MOVE   (((flags&COPYFLAG_MOVE)!=0)||MY_DELETE)
#define MY_FAIL   ((flags&COPYFLAG_FAIL)!=0)
#define MY_ROSRC  ((flags&COPYFLAG_ROSRC)!=0)
#define MY_RODST  ((flags&COPYFLAG_RODST)!=0)

#define MY_STATISTIC  ((flags&COPYFLAG_STATISTIC)!=0)
#define MY_ACCESS     ((flags&COPYFLAG_ACCESS)!=0)
#define MY_COPY_SECURITY \
if((ret==retOk)&&(rDestW[wcslen(rDestW)-1]==L'\\')) MyCopySecurityW(rSrcW,rDestW,MY_ACCESS,TRUE);

DWORD CopyCallback(wchar_t *src,wchar_t *dest,DWORD flags)
{
  BOOL bExist=FALSE;
  WIN32_FIND_DATAW findin,findout; HANDLE hFind;
  BYTE buff[64*1024]; DWORD dReaded,dWritten;
  hFind=FindFirstFileW(dest,&findout);
  if(hFind!=INVALID_HANDLE_VALUE)
  {
    bExist=TRUE;
    FindClose(hFind);
    if(MY_RODST && (findout.dwFileAttributes&FILE_ATTRIBUTE_READONLY))
      SetFileAttributesW(dest,findout.dwFileAttributes&(~FILE_ATTRIBUTE_READONLY));
  }
  hFind=FindFirstFileW(src,&findin);
  if(hFind==INVALID_HANDLE_VALUE)
  {
    Log(L"FindFirstFileW",src,NULL);
    return retError;
  }
  FindClose(hFind);
  if(MY_MOVE && MY_ROSRC && (findin.dwFileAttributes&FILE_ATTRIBUTE_READONLY))
    SetFileAttributesW(src,findin.dwFileAttributes&(~FILE_ATTRIBUTE_READONLY));
  switch(flags&(COPYFLAG_OVERWRITE|COPYFLAG_APPEND|COPYFLAG_SKIP|COPYFLAG_REFRESH))
  {
    case COPYFLAG_REFRESH:
      if(bExist&&(CompareFileTime(&(findin.ftLastWriteTime),&(findout.ftLastWriteTime))!=1))
        break;
    case COPYFLAG_OVERWRITE:
      return MyCopyFileW(src,dest,FALSE,MY_ACCESS,MY_MOVE);
    case COPYFLAG_APPEND:
      if(bExist)
      {
        HANDLE hIn=INVALID_HANDLE_VALUE,hOut=INVALID_HANDLE_VALUE;
        unsigned long long subadd=0; long dist; DWORD Ok=FALSE;
        if(UpdateInfo(src,dest)) return retStop;
        while(!Ok)
        {
          hIn=CreateFileW(src,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,FILE_FLAG_SEQUENTIAL_SCAN,NULL);
          if(hIn==INVALID_HANDLE_VALUE)
          {
            Log(L"CreateFileW",src,NULL);
            break;
          }
          hOut=CreateFileW(dest,GENERIC_WRITE,FILE_SHARE_READ,NULL,OPEN_EXISTING,0,NULL);
          if(hIn==INVALID_HANDLE_VALUE)
          {
            Log(L"CreateFileW",dest,NULL);
            break;
          }
          dist=0;
          if((SetFilePointer(hOut,0,&dist,FILE_END)!=0xFFFFFFFF)||(GetLastError()==NO_ERROR))
          {
            Ok=TRUE;
            do
            {
              if(UpdatePosInfo(0,subadd)) break;
              if(!ReadFile(hIn,buff,sizeof(buff),&dReaded,NULL))
              {
                Log(L"ReadFile",src,NULL); Ok=FALSE; break;
              }
              subadd+=dReaded;
              if(!WriteFile(hOut,buff,dReaded,&dWritten,NULL))
              {
                Log(L"WriteFile",dest,NULL); Ok=FALSE; break;
              }
            } while(dReaded==sizeof(buff));
          }
          else
          {
            Log(L"SetFilePointer",NULL,NULL);
            break;
          }
        }
        if(hIn!=INVALID_HANDLE_VALUE) CloseHandle(hIn);
        if(hOut!=INVALID_HANDLE_VALUE) CloseHandle(hOut);
        if(!Ok) return retError;
        if(MY_MOVE)
          if(!DeleteFileW(src))
          {
            Log(L"DeleteFileW",src,NULL);
            return retError;
          }
        break;
      }
    case COPYFLAG_SKIP:
      if(!bExist)
        return MyCopyFileW(src,dest,TRUE,MY_ACCESS,MY_MOVE);
      break;
  }
  return retOk;
}

DWORD RecurseDir(wchar_t *dir,wchar_t *dest,DWORD flags)
{
  WIN32_FIND_DATAW find; HANDLE hFind; DWORD ret=retOk;
  wchar_t findmaskW[MAX_PATH];
  wchar_t rSrcW[MAX_PATH],rDestW[MAX_PATH]=L"";
  swprintf(findmaskW,L"%s\\*",dir);
  hFind=FindFirstFileW(findmaskW,&find);
  if(hFind!=INVALID_HANDLE_VALUE)
  {
    do
    {
      if(find.dwFileAttributes&FILE_ATTRIBUTE_DIRECTORY)
      {
        if(wcscmp(find.cFileName,L".")&&wcscmp(find.cFileName,L".."))
        {
          swprintf(rSrcW,L"%s\\%s",dir,find.cFileName);
          if(!(find.dwFileAttributes&FILE_ATTRIBUTE_REPARSE_POINT))
          {
            if(!MY_DELETE)
            {
              if(dest[wcslen(dest)-1]==L'\\')
              {
                swprintf(rDestW,L"%s%s\\",dest,find.cFileName);
                CreateDirectoryW(rDestW,NULL);
              }
              else
                swprintf(rDestW,L"%s",dest);
            }
            ret=RecurseDir(rSrcW,rDestW,flags);
            MY_COPY_SECURITY
            if(((ret==retError) && MY_FAIL) || (ret==retStop))
              break;
          }
          if(MY_MOVE) MyRemoveDirectoryW(rSrcW);
        }
      }
      else
      {
        swprintf(rSrcW,L"%s\\%s",dir,find.cFileName);
        if(!MY_DELETE)
        {
          if(dest[wcslen(dest)-1]==L'\\')
            swprintf(rDestW,L"%s%s",dest,find.cFileName);
          else
            swprintf(rDestW,L"%s",dest);
        }
        ret=CopyCallback(rSrcW,rDestW,flags);
        if(MY_STATISTIC)
          if(UpdatePosInfo(MyGetFileSize(MY_MOVE?rDestW:rSrcW),0))
            ret=retStop;
        if(((ret==retError) && MY_FAIL) || (ret==retStop))
          break;
      }
    } while(FindNextFileW(hFind,&find));
    FindClose(hFind);
  }
  return ret;
}

void ScanDir(wchar_t *dir,unsigned long long *size)
{
  WIN32_FIND_DATAW find; HANDLE hFind;
  wchar_t findmaskW[MAX_PATH];
  wchar_t rSrcW[MAX_PATH];
  swprintf(findmaskW,L"%s\\*",dir);
  hFind=FindFirstFileW(findmaskW,&find);
  if(hFind!=INVALID_HANDLE_VALUE)
  {
    do
    {
      if(find.dwFileAttributes&FILE_ATTRIBUTE_DIRECTORY)
      {
        if(wcscmp(find.cFileName,L".")&&wcscmp(find.cFileName,L".."))
        {
          swprintf(rSrcW,L"%s\\%s",dir,find.cFileName);
          ScanDir(rSrcW,size);
        }
      }
      else
        *size+=(unsigned long long)find.nFileSizeLow+(unsigned long long)find.nFileSizeHigh*4294967296ULL;
    } while(FindNextFileW(hFind,&find));
    FindClose(hFind);
  }
}

unsigned long long MyGetFileSize(wchar_t *file)
{
  WIN32_FIND_DATAW find; HANDLE hFind;
  hFind=FindFirstFileW(file,&find);
  if(hFind!=INVALID_HANDLE_VALUE)
  {
    FindClose(hFind);
    return (unsigned long long)find.nFileSizeLow+(unsigned long long)find.nFileSizeHigh*4294967296ULL;
  }
  return 0;
}

DWORD WINAPI CopyThread(LPVOID lpvThreadParm)
{
  DWORD count,i;
  struct FileRec *files;
  struct FileRec *src;
  struct FileRec *dest;
  DWORD flags,type=INFOTYPE_COPY,ret;
  unsigned long long size;

  wchar_t rSrcW[MAX_PATH],rDestW[MAX_PATH];
  
  flags=((struct FileRecs *)lpvThreadParm)->flags;
  count=((struct FileRecs *)lpvThreadParm)->count;
  files=((struct FileRecs *)lpvThreadParm)->files;
  src=(files+count-2);
  dest=(files+count-1);

  //rename stuff start
  if((!MY_DELETE)&&(MY_MOVE)&&(count==3)&&(dest->name[wcslen(dest->name)-1]!=L'\\')&&(files[0].type&FILETYPE_FOLDER))
  {
    swprintf(rSrcW,L"%s%s",src->name,files[0].name);
    MyCopyFileW(rSrcW,dest->name,TRUE,FALSE,TRUE);
    free(lpvThreadParm);
    free(files);
    return 0;
  }
  //rename stuff end

  //add to info list
  if(MY_MOVE) type++;
  if(MY_DELETE) type++;
  AddInfo(type,src->name,MY_STATISTIC);

  //scan dirs for size
  if(MY_STATISTIC)
  {
    size=0;
    for(i=0;i<count-2;i++)
    {
      swprintf(rSrcW,L"%s%s",src->name,files[i].name);
      if(files[i].type&FILETYPE_FILE)
      {
        size+=MyGetFileSize(rSrcW);
      }
      else if(files[i].type&FILETYPE_FOLDER)
      {
        if(!(files[i].type&FILETYPE_JUNCTION))
          ScanDir(rSrcW,&size);
      }
    }
    SetStartInfo(size);
  }

  rDestW[0]=0; //for delete
  for(i=0;i<count-2;i++)
  {
    swprintf(rSrcW,L"%s%s",src->name,files[i].name);
    if(files[i].type&FILETYPE_FILE)
    {
      if(!MY_DELETE)
      {
        if(dest->name[wcslen(dest->name)-1]==L'\\')
          swprintf(rDestW,L"%s%s",dest->name,files[i].name);
        else
          swprintf(rDestW,L"%s",dest->name);
      }
      ret=CopyCallback(rSrcW,rDestW,flags);
      if(MY_STATISTIC)
        if(UpdatePosInfo(MyGetFileSize(MY_MOVE?rDestW:rSrcW),0))
          ret=retStop;
      if(((ret==retError) && MY_FAIL) || (ret==retStop))
        break;
    }
    else if(files[i].type&FILETYPE_FOLDER)
    {
      if(!(files[i].type&FILETYPE_JUNCTION))
      {
        if(!MY_DELETE)
        {
          if(dest->name[wcslen(dest->name)-1]==L'\\')
          {
            swprintf(rDestW,L"%s%s\\",dest->name,files[i].name);
            CreateDirectoryW(rDestW,NULL);
          }
          else
            swprintf(rDestW,L"%s",dest->name);
        }
        ret=RecurseDir(rSrcW,rDestW,flags);
        MY_COPY_SECURITY
        if(((ret==retError) && MY_FAIL) || (ret==retStop))
          break;
      }
      if(MY_MOVE) MyRemoveDirectoryW(rSrcW);
    }
  }
  DelInfo();
  free(lpvThreadParm);
  free(files);
  return 0;
}

#undef MY_DELETE
#undef MY_MOVE
#undef MY_FAIL
#undef MY_ROSRC
#undef MY_RODST
#undef MY_STATISTIC
#undef MY_ACCESS
#undef MY_COPY_SECURITY

//ASCII thread. Build structures for CopyThread from params.

DWORD WINAPI PrepareThread(LPVOID lpvThreadParm)
{
  struct StrRec *rec=lpvThreadParm;
  HANDLE hThread;
  DWORD ThreadID;
  HANDLE hFind; WIN32_FIND_DATAW find;
  HANDLE client_token;
  DWORD dCount=0;
  struct FileRec *files=NULL;
  struct FileRecs *sfiles=NULL;
  wchar_t SrcW[MAX_PATH]; wchar_t *slash;

  memset(SrcW,0,sizeof(SrcW));

  slash=wcsrchr(rec->from,'\\');
  if((!slash)||(slash<wcsrchr(rec->from,'/')))
    slash=wcsrchr(rec->from,'/');
  if(slash)
    wcsncpy(SrcW,rec->from,slash-rec->from+1);
  hFind=FindFirstFileW(rec->from,&find);
  if(hFind!=INVALID_HANDLE_VALUE)
  {
    do
    {
      if(find.dwFileAttributes&FILE_ATTRIBUTE_DIRECTORY)
      {
        if(wcscmp(find.cFileName,L".")&&wcscmp(find.cFileName,L".."))
        {
          files=(struct FileRec *)realloc(files,sizeof(struct FileRec)*(dCount+1));
          files[dCount].type=FILETYPE_FOLDER;
          if(find.dwFileAttributes&FILE_ATTRIBUTE_REPARSE_POINT)
            files[dCount].type|=FILETYPE_JUNCTION;
          swprintf(files[dCount].name,L"%s",find.cFileName);
          dCount++;
        }
      }
      else
      {
        files=(struct FileRec *)realloc(files,sizeof(struct FileRec)*(dCount+1));
        files[dCount].type=FILETYPE_FILE;
        swprintf(files[dCount].name,L"%s",find.cFileName);
        dCount++;
      }
    } while(FindNextFileW(hFind,&find));
    FindClose(hFind);
  }
  if(dCount)
  {
    files=(struct FileRec *)realloc(files,sizeof(struct FileRec)*(dCount+2));
    files[dCount].type=FILETYPE_FOLDER;
    swprintf(files[dCount].name,L"%s",SrcW);
    dCount++;

    files[dCount].type=FILETYPE_FOLDER;
    swprintf(files[dCount].name,L"%s",rec->to);
    dCount++;

    sfiles=(struct FileRecs *)malloc(sizeof(struct FileRecs));
    sfiles->flags=rec->flags;
    sfiles->count=dCount;
    sfiles->files=files;
    hThread=CreateThread(NULL,0,CopyThread,sfiles,CREATE_SUSPENDED,&ThreadID);
    SetThreadPriority(hThread,THREAD_PRIORITY_IDLE);
    //Security
    if(OpenThreadToken(GetCurrentThread(),TOKEN_ALL_ACCESS,FALSE,&client_token))
    {
      SetThreadToken(&hThread,client_token);
      CloseHandle(client_token);
    }
    ResumeThread(hThread);
    CloseHandle(hThread);
    files=NULL;
    sfiles=NULL;
  }
  free(lpvThreadParm);
  return 0;
}