DWORD __stdcall RtlNtStatusToDosError(long Status);

long __stdcall ZwCreateToken(
  PHANDLE TokenHandle,
  ACCESS_MASK DesiredAccess,
  POBJECT_ATTRIBUTES ObjectAttributes,
  TOKEN_TYPE Type,
  PLUID AuthenticationId,
  PLARGE_INTEGER ExpirationTime,
  PTOKEN_USER User,
  PTOKEN_GROUPS Groups,
  PTOKEN_PRIVILEGES Privileges,
  TOKEN_OWNER *Owner,
  TOKEN_PRIMARY_GROUP *PrimaryGroup,
  TOKEN_DEFAULT_DACL *DefaultDacl,
  PTOKEN_SOURCE Source
);

static void EnablePrivilege(const char *name)
{
  TOKEN_PRIVILEGES priv;
  HANDLE token;
  priv.PrivilegeCount=1;
  priv.Privileges[0].Attributes=SE_PRIVILEGE_ENABLED;

  if(!LookupPrivilegeValue(0,name,&(priv.Privileges[0].Luid)))
    LogSys(L"LookupPrivilegeValue",NULL,NULL);
  else
  {
    if(!OpenProcessToken(GetCurrentProcess(),TOKEN_ADJUST_PRIVILEGES, &token))
      LogSys(L"OpenProcessToken",NULL,NULL);
    else
    {
      AdjustTokenPrivileges(token, FALSE, &priv, sizeof(priv), 0, 0);
      if(GetLastError()!=ERROR_SUCCESS)
        LogSys(L"AdjustTokenPrivileges",NULL,NULL);
      CloseHandle(token);
    }
  }
}

static void *DefaultTokenInformation(HANDLE token,TOKEN_INFORMATION_CLASS tic)
{
  unsigned long size; unsigned char *Result=NULL;
  GetTokenInformation(token,tic,0,0,&size);
  Result=(unsigned char *)malloc(size);
  if(Result)
  {
    if(!GetTokenInformation(token,tic,Result,size,&size))
      LogSys(L"GetTokenInformation",NULL,NULL);
  }
  return Result;
}

static void PipeToThread(HANDLE pipe,HANDLE thread)
{
  HANDLE pipe_token,new_token;
  SECURITY_QUALITY_OF_SERVICE sqos={sizeof sqos,SecurityImpersonation,SECURITY_STATIC_TRACKING,FALSE};
  OBJECT_ATTRIBUTES oa={sizeof(oa),0,0,0,0,&sqos};
  TOKEN_STATISTICS *stats=NULL;
  PTOKEN_USER user=NULL; PTOKEN_GROUPS groups=NULL; PTOKEN_PRIVILEGES privileges=NULL;
  TOKEN_OWNER *owner=NULL; TOKEN_PRIMARY_GROUP *group=NULL; TOKEN_DEFAULT_DACL *dacl=NULL;
  long res; BOOL CanOpenThread;

  if(!ImpersonateNamedPipeClient(pipe))
    LogSys(L"ImpersonateNamedPipeClient",NULL,NULL);
  else
  {
    CanOpenThread=OpenThreadToken(GetCurrentThread(),TOKEN_ALL_ACCESS,FALSE,&pipe_token);
    if(!CanOpenThread)
      LogSys(L"OpenThreadToken",NULL,NULL);
    if(!RevertToSelf())
      LogSys(L"RevertToSelf",NULL,NULL);
    if(CanOpenThread)
    {
      if(LogonAs)
      {
        stats=(TOKEN_STATISTICS *)DefaultTokenInformation(pipe_token,TokenStatistics);
        if(!stats) goto cleanup;
        user=(PTOKEN_USER)DefaultTokenInformation(pipe_token,TokenUser);
        if(!user) goto cleanup;
        groups=(PTOKEN_GROUPS)DefaultTokenInformation(pipe_token,TokenGroups);
        if(!groups) goto cleanup;
        privileges=(PTOKEN_PRIVILEGES)DefaultTokenInformation(pipe_token,TokenPrivileges);
        if(!privileges) goto cleanup;
        owner=(TOKEN_OWNER *)DefaultTokenInformation(pipe_token,TokenOwner);
        if(!owner) goto cleanup;
        group=(TOKEN_PRIMARY_GROUP *)DefaultTokenInformation(pipe_token,TokenPrimaryGroup);
        if(!group) goto cleanup;
        dacl=(TOKEN_DEFAULT_DACL *)DefaultTokenInformation(pipe_token,TokenDefaultDacl);
        if(!dacl) goto cleanup;
        res=ZwCreateToken(&new_token,TOKEN_ALL_ACCESS,&oa,TokenImpersonation,&BCopyID,&(stats->ExpirationTime),user,groups,privileges,owner,group,dacl,&NetworkSource);
        if(res)
        {
          DWORD doserr=RtlNtStatusToDosError(res);
          SetLastError(doserr);
          Log(L"ZwCreateToken",NULL,NULL);
        }
        else
        {
          if(!SetThreadToken(&thread,new_token))
            Log(L"SetThreadToken",NULL,NULL);
          CloseHandle(new_token);
        }
cleanup:
        if(stats) free(stats);
        if(user) free(user);
        if(groups) free(groups);
        if(privileges) free(privileges);
        if(owner) free(owner);
        if(group) free(group);
        if(dacl) free(dacl);
        CloseHandle(pipe_token);
      }
      else
      {
        if(!SetThreadToken(&thread,pipe_token))
          Log(L"SetThreadToken",NULL,NULL);
        CloseHandle(pipe_token);
      }
    }
  }
}
