#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "..\plugin.hpp"
#include "memory.h"
#include "bcplugin.h"
#include "bcCommon.h"

PluginStartupInfo Info;
FARSTANDARDFUNCTIONS FSF;
char PluginRootKey[80];
BOOL IsOldFAR=TRUE;

struct Options{
  int  CopyType;
  BOOL CopyHistory;
  BOOL CopyROSrc;
  BOOL CopyRODest;
  BOOL CopyAbort;
  BOOL CopyFullInfo;
  BOOL CopyAccess;
  BOOL DeleteRO;
  BOOL DeleteAbort;
} Opt={2,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE};

#define SHOW_IN_VIEWER 1
#define SHOW_IN_EDITOR 2
#define SHOW_IN_CONFIG 4

#define INFO_MENU_ALT_0 1
#define INFO_MENU_ALT_1 2

struct PlugOptions{
  DWORD AutoStart;
  DWORD ShowMenu;
  DWORD InfoMenu;
  DWORD ExpandVars;
  BOOL CheckPassive;
} PlgOpt={1,SHOW_IN_VIEWER|SHOW_IN_EDITOR|SHOW_IN_CONFIG,0,0,TRUE};

char *GetMsg(int MsgId)
{
  return(Info.GetMsg(Info.ModuleNumber, MsgId));
};

#include "bcplugmix.cpp"
#include "bcopy_eject.cpp"
char MsgCopy[512];
char MsgMove[512];
#include "bcplugdialogproc.cpp"
#include "bcconfig.cpp"

void WINAPI _export SetStartupInfo(const struct PluginStartupInfo *Info)
{
  memset(&::Info, 0, sizeof(::Info));
  memmove(&::Info, Info, (Info->StructSize > (int)sizeof(::Info))?sizeof(::Info):Info->StructSize);
  if(Info->StructSize > FAR165_INFO_SIZE)
  {
    IsOldFAR=FALSE;
    ::FSF=*Info->FSF;
    ::Info.FSF=&::FSF;

    strcpy(PluginRootKey,Info->RootKey);
    strcat(PluginRootKey,"\\BCopy");

    HKEY hKey;
    DWORD Type;
    DWORD DataSize=0;
    if ((RegOpenKeyEx(HKEY_CURRENT_USER,PluginRootKey,0,
           KEY_QUERY_VALUE,&hKey))==ERROR_SUCCESS)
    {
      //tech options
      DataSize=sizeof(PlgOpt.AutoStart);
      RegQueryValueEx(hKey,"TechAutoStart",0,&Type,(LPBYTE)&PlgOpt.AutoStart,&DataSize);
      DataSize=sizeof(PlgOpt.ShowMenu);
      RegQueryValueEx(hKey,"TechShowMenu",0,&Type,(LPBYTE)&PlgOpt.ShowMenu,&DataSize);
      DataSize=sizeof(PlgOpt.InfoMenu);
      RegQueryValueEx(hKey,"TechInfoMenu",0,&Type,(LPBYTE)&PlgOpt.InfoMenu,&DataSize);
      DataSize=sizeof(PlgOpt.ExpandVars);
      RegQueryValueEx(hKey,"TechExpandVars",0,&Type,(LPBYTE)&PlgOpt.ExpandVars,&DataSize);
      DataSize=sizeof(PlgOpt.CheckPassive);
      RegQueryValueEx(hKey,"TechCheckPassive",0,&Type,(LPBYTE)&PlgOpt.CheckPassive,&DataSize);
      //copy options
      DataSize=sizeof(Opt.CopyType);
      RegQueryValueEx(hKey,"CopyType",0,&Type,(LPBYTE)&Opt.CopyType,&DataSize);
      DataSize=sizeof(Opt.CopyHistory);
      RegQueryValueEx(hKey,"CopyHistory",0,&Type,(LPBYTE)&Opt.CopyHistory,&DataSize);
      DataSize=sizeof(Opt.CopyROSrc);
      RegQueryValueEx(hKey,"CopyROSrc",0,&Type,(LPBYTE)&Opt.CopyROSrc,&DataSize);
      DataSize=sizeof(Opt.CopyRODest);
      RegQueryValueEx(hKey,"CopyRODest",0,&Type,(LPBYTE)&Opt.CopyRODest,&DataSize);
      DataSize=sizeof(Opt.CopyAbort);
      RegQueryValueEx(hKey,"CopyAbort",0,&Type,(LPBYTE)&Opt.CopyAbort,&DataSize);
      DataSize=sizeof(Opt.CopyFullInfo);
      RegQueryValueEx(hKey,"CopyFullInfo",0,&Type,(LPBYTE)&Opt.CopyFullInfo,&DataSize);
      DataSize=sizeof(Opt.CopyAccess);
      RegQueryValueEx(hKey,"CopyAccess",0,&Type,(LPBYTE)&Opt.CopyAccess,&DataSize);
      //delete options
      DataSize=sizeof(Opt.DeleteRO);
      RegQueryValueEx(hKey,"DeleteRO",0,&Type,(LPBYTE)&Opt.DeleteRO,&DataSize);
      DataSize=sizeof(Opt.DeleteAbort);
      RegQueryValueEx(hKey,"DeleteAbort",0,&Type,(LPBYTE)&Opt.DeleteAbort,&DataSize);
      RegCloseKey(hKey);
    }
    if((Opt.CopyType>3)||(Opt.CopyType<0))
      Opt.CopyType=2;
  }
}

void WINAPI _export GetPluginInfo(struct PluginInfo *Info)
{
  if(!IsOldFAR)
  {
    Info->StructSize=sizeof(*Info);
    Info->Flags=0;
    if(PlgOpt.ShowMenu&SHOW_IN_VIEWER)
      Info->Flags|=PF_VIEWER;
    if(PlgOpt.ShowMenu&SHOW_IN_EDITOR)
      Info->Flags|=PF_EDITOR;
    Info->DiskMenuStringsNumber=0;
    static char *PluginMenuStrings[1];
    PluginMenuStrings[0]=GetMsg(mName);
    Info->PluginMenuStrings=PluginMenuStrings;
    Info->PluginMenuStringsNumber=sizeof(PluginMenuStrings)/sizeof(PluginMenuStrings[0]);
    Info->PluginConfigStrings=PluginMenuStrings;
    Info->PluginConfigStringsNumber=0;
    if(PlgOpt.ShowMenu&SHOW_IN_CONFIG)
      Info->PluginConfigStringsNumber=sizeof(PluginMenuStrings)/sizeof(PluginMenuStrings[0]);
  }
}

int WINAPI _export Configure(int ItemNumber)
{
  switch(ItemNumber)
  {
    case 0:
      return(Config());
  }
  return(FALSE);
}

static void SendToPipe(DWORD *send,DWORD sendsize)
{
  HANDLE hPipe=CreateFile(PIPE_NAME,GENERIC_READ|GENERIC_WRITE,0,NULL,OPEN_EXISTING,0,NULL);
  if(hPipe!=INVALID_HANDLE_VALUE)
  {
    DWORD cbWritten; 
    WriteFile(hPipe,send,sendsize,&cbWritten,NULL);
    CloseHandle(hPipe);
  }
}

static BOOL CheckPipe(void)
{
  HANDLE hPipe=CreateFile(PIPE_NAME,GENERIC_READ|GENERIC_WRITE,0,NULL,OPEN_EXISTING,0,NULL);
  if(hPipe!=INVALID_HANDLE_VALUE)
  {
    CloseHandle(hPipe);
    return TRUE;
  }
  return FALSE;
}

static BOOL CheckPipeEx(void)
{
  BOOL Result=CheckPipe();
  if(!Result)
  {
    if(PlgOpt.AutoStart)
    {
      SC_HANDLE ServiceControlHandle;
      SC_HANDLE SCManagerHandle;
      SERVICE_STATUS ServiceStatus;

      SCManagerHandle=OpenSCManager(NULL,NULL,SC_MANAGER_ALL_ACCESS);
      if(SCManagerHandle>0)
      {
        ServiceControlHandle=OpenService(SCManagerHandle,SVC_NAME,SERVICE_ALL_ACCESS);
        if(ServiceControlHandle>0)
        {
          if(StartService(ServiceControlHandle,0,NULL))
          {
            while(QueryServiceStatus(ServiceControlHandle,&ServiceStatus))
            {
              if(ServiceStatus.dwCurrentState==SERVICE_START_PENDING)
                Sleep(200);
              else
                break;
            }
            if(ServiceStatus.dwCurrentState==SERVICE_RUNNING)
              Result=CheckPipe();
          }
          CloseServiceHandle(ServiceControlHandle);
        }
        CloseServiceHandle(SCManagerHandle);
      }
    }
  }
  return Result;
}

#define MY_COMPINE_PATH \
FSF.AddEndSlash(oem_buff); \
sprintf(res_buff,"%s%s",oem_buff,path+3); \
sprintf(path,"%s",res_buff);
static void UNCPath(char *path)
{
  char buff[MAX_PATH],oem_buff[MAX_PATH],res_buff[MAX_PATH],device[3]=" :"; DWORD size=sizeof(buff);
  if((strlen(path)>2)&&(path[1]==':'))
  {
    SetFileApisToANSI();
    device[0]=path[0];
    if(WNetGetConnection(device,buff,&size)==NO_ERROR)
    {
      CharToOem(buff,oem_buff);
      MY_COMPINE_PATH
    }
    else if(QueryDosDevice(device,buff,size))
      if(!strncmp(buff,"\\??\\",4))
      {
        CharToOem(buff+4,oem_buff);
        MY_COMPINE_PATH
      }
    SetFileApisToOEM();
  }
}
#undef MY_COMPINE_PATH

static int NumberType(int num)
{
  int Result=0,mod;
  mod=num%100;
  if((mod>20)||(mod<5))
  {
    mod=num%10;
    if((mod<5)&&(mod>0)) Result--;
    if(mod==1) Result--;
  }
  return Result;
}

static const char *GetRealName(const WIN32_FIND_DATAA *src)
{
  WIN32_FIND_DATAA find,find_ok; HANDLE hFind; BOOL Res;
  hFind=FindFirstFileA(src->cFileName,&find);
  if(hFind!=INVALID_HANDLE_VALUE)
  {
    memcpy(&find_ok,&find,sizeof(find));
    Res=FindNextFileA(hFind,&find);
    FindClose(hFind);
    if((!Res)&&(!stricmp(src->cAlternateFileName,find_ok.cAlternateFileName)))
    {
      return src->cFileName;
    }
  }
  if(src->cAlternateFileName[0])
  {
    hFind=FindFirstFileA(src->cAlternateFileName,&find);
    if(hFind!=INVALID_HANDLE_VALUE)
    {
      FindClose(hFind);
      return src->cAlternateFileName;
    }
  }
  return NULL;  
}

HANDLE WINAPI _export OpenPlugin(int OpenFrom,int Item)
{
//  int MenuCode=mMenuInfo-mMenuCopy;
  int MenuCode=-1;
  FarMenuItem MenuItems[mMenuInfo-mMenuCopy+1];
  unsigned int MenuItemsSize=mMenuInfo-mMenuCopy+1;
  memset(MenuItems,0,sizeof(MenuItems));
  int Msgs[]={mMenuCopy,mMenuMove,mMenuDelete,mMenuView,mMenuEdit,mMenuSep1,mMenuEject,mMenuReject,mMenuSep2,mMenuConfig,mMenuInfo};
  if(OpenFrom!=OPEN_PLUGINSMENU)
  {
    MenuItemsSize=mMenuInfo-mMenuSep1;
    memcpy(Msgs,Msgs+mMenuEject-mMenuCopy,sizeof(Msgs[0])*MenuItemsSize);
  }
  for(unsigned int i=0;i<MenuItemsSize;++i)
  {
    MenuItems[i].Selected=MenuItems[i].Checked=MenuItems[i].Separator=0;
    sprintf(MenuItems[i].Text, "%s", GetMsg(Msgs[i])); // Text in menu
    if(MenuItems[i].Text[0]=='-') MenuItems[i].Separator=TRUE;
  };

  // First item is selected
  MenuItems[0].Selected=TRUE;
  // Show menu
  MenuCode=Info.Menu(Info.ModuleNumber,-1,-1,0,FMENU_AUTOHIGHLIGHT|FMENU_WRAPMODE,
                         NULL,NULL,"Contents",NULL,NULL,
                         MenuItems,MenuItemsSize);
  if((OpenFrom!=OPEN_PLUGINSMENU)&&(MenuCode>=0))
    MenuCode+=mMenuEject-mMenuCopy;
  //prepare source dir
  PanelInfo PInfo;
  Info.Control(INVALID_HANDLE_VALUE,FCTL_GETPANELINFO,&PInfo);

  if((MenuCode<5) && (MenuCode>=0))
  {
    if((PInfo.SelectedItemsNumber>0)&&(!PInfo.Plugin))
    {
      char SrcA[MAX_PATH]; wchar_t SrcW[MAX_PATH];
      strcpy(SrcA, PInfo.CurDir);
      UNCPath(SrcA);
      FSF.AddEndSlash(SrcA);
      MultiByteToWideChar(CP_OEMCP,0,SrcA,-1,SrcW,MAX_PATH);

      DWORD *send=NULL;
      DWORD sendsize=sizeof(DWORD)*3+sizeof(FileRec)*(PInfo.SelectedItemsNumber+2);
      if(MenuCode<3)
        send=(DWORD *)malloc(sendsize);

      switch(MenuCode)
      {
        case 0:
        case 1:
        {
          if(!send) break;
          if(!CheckPipeEx()) break;
          PanelInfo PAnotherInfo;
          Info.Control(INVALID_HANDLE_VALUE,FCTL_GETANOTHERPANELINFO,&PAnotherInfo);
          if(PlgOpt.CheckPassive&&PAnotherInfo.Plugin)
            break;

          char DestA[MAX_PATH]; wchar_t DestW[MAX_PATH];
          if(PAnotherInfo.Plugin)
            strcpy(DestA,"");
          else
          {
            strcpy(DestA,PAnotherInfo.CurDir);
            UNCPath(DestA);
            FSF.AddEndSlash(DestA);
          }

          //Show dialog
          /*
            0000000000111111111122222222223333333333444444444455555555556666666666777777
            0123456789012345678901234567890123456789012345678901234567890123456789012345
          00                                                                            00
          01    Copy ͻ   01
          02    Copy "bcopy.cpp" to                                                   02
          03    \WINNT\Profiles\Administrator\Personal\!Far_plugs\bc\doc\reversi\ |   03
          04   Ķ   04
          05    ( ) Overwrite all existing files                                      05
          06    ( ) Append to all existing files                                      06
          07    ( ) Skip all existing files                                           07
          08    ( ) Refresh old files                                                 08
          09                                                                          09
          10    [ ] Remove source files                                               10
          11    [ ] Copy access rights                                                11
          12    [ ] Clear R/O attribute on &source                                    12
          13    [ ] Clear R/O attribute on &destination                               13
          14    [ ] Abort on error                                                    14
          15   Ķ   15
          16                    [ Copy ]                [ Cancel ]                    16
          17   ͼ   17
          18                                                                            18
            0000000000111111111122222222223333333333444444444455555555556666666666777777
            0123456789012345678901234567890123456789012345678901234567890123456789012345
          */
          char *CopyHistoryName="Copy";
          char *BCopyCopyHistoryName="BCopyCopy";
          static struct InitDialogItem InitDlg[]={
          /* 0*/  {DI_DOUBLEBOX,3,1,72,17,0,0,0,0,(char *)mCpyDlgCopyTitle},
          /* 1*/  {DI_TEXT,5,2,0,0,0,0,0,0,""},
          /* 2*/  {DI_EDIT,5,3,70,0,1,(DWORD)BCopyCopyHistoryName,DIF_HISTORY,0,""},
          /* 3*/  {DI_TEXT,5,4,0,0,0,0,DIF_BOXCOLOR|DIF_SEPARATOR,0,""},
          /* 4*/  {DI_RADIOBUTTON,5,5,0,0,0,0,DIF_GROUP,0,(char *)mCpyDlgOpt1},
          /* 5*/  {DI_RADIOBUTTON,5,6,0,0,0,0,0,0,(char *)mCpyDlgOpt2},
          /* 6*/  {DI_RADIOBUTTON,5,7,0,0,0,0,0,0,(char *)mCpyDlgOpt3},
          /* 7*/  {DI_RADIOBUTTON,5,8,0,0,0,0,0,0,(char *)mCpyDlgOpt4},
          /* 8*/  {DI_CHECKBOX,5,10,0,0,0,0,0,0,(char *)mCpyDlgChk1},
          /* 9*/  {DI_CHECKBOX,5,11,0,0,0,0,0,0,(char *)mCpyDlgChkAccess},
          /*10*/  {DI_CHECKBOX,5,12,0,0,0,0,DIF_DISABLE,0,(char *)mCpyDlgChk2},
          /*11*/  {DI_CHECKBOX,5,13,0,0,0,0,DIF_DISABLE,0,(char *)mCpyDlgChk3},
          /*12*/  {DI_CHECKBOX,5,14,0,0,0,0,0,0,(char *)mCpyDlgChk4},
          /*13*/  {DI_TEXT,5,15,0,0,0,0,DIF_BOXCOLOR|DIF_SEPARATOR,0,""},
          /*14*/  {DI_BUTTON,0,16,0,0,0,0,DIF_CENTERGROUP,1,(char *)mCpyDlgOkCopy},
          /*15*/  {DI_BUTTON,0,16,0,0,0,0,DIF_CENTERGROUP,0,(char *)mCpyDlgCancel}
          };

          struct FarDialogItem DialogItems[sizeof(InitDlg)/sizeof(InitDlg[0])];
          InitDialogItems(InitDlg,DialogItems,sizeof(InitDlg)/sizeof(InitDlg[0]));
          if(Opt.CopyHistory)
            DialogItems[2].History=CopyHistoryName;
          DialogItems[4+Opt.CopyType].Selected=1;
          DialogItems[8].Selected=MenuCode;
          DialogItems[9].Selected=Opt.CopyAccess;
          DialogItems[10].Selected=Opt.CopyROSrc;
          DialogItems[11].Selected=Opt.CopyRODest;
          DialogItems[12].Selected=Opt.CopyAbort;
          if(PInfo.SelectedItemsNumber>1)
          {
            sprintf(MsgCopy,GetMsg(mCpyDlgCopyToN+NumberType(PInfo.SelectedItemsNumber)),PInfo.SelectedItemsNumber);
            sprintf(MsgMove,GetMsg(mCpyDlgMoveToN+NumberType(PInfo.SelectedItemsNumber)),PInfo.SelectedItemsNumber);
          }
          else
          {
            sprintf(MsgCopy,GetMsg(mCpyDlgCopyTo1),PInfo.SelectedItems[0].FindData.cFileName);
            sprintf(MsgMove,GetMsg(mCpyDlgMoveTo1),PInfo.SelectedItems[0].FindData.cFileName);
          }
          sprintf(DialogItems[1].Data,MsgCopy);
          sprintf(DialogItems[2].Data,"%s",DestA);
          int DlgCode=Info.DialogEx(Info.ModuleNumber,-1,-1,76,19,"Copy",DialogItems,(sizeof(InitDlg)/sizeof(InitDlg[0])),0,0,CopyDialogProc,0);
          if(DlgCode==14)
          {
            if(PlgOpt.ExpandVars)
            {
              char exp[512]; DWORD res;
              res=ExpandEnvironmentStrings(DialogItems[2].Data,exp,sizeof(exp));
              if(res&&(res<=sizeof(exp)))
                strcpy(DialogItems[2].Data,exp);
            }
            char *filename;
            filename=strrchr(DialogItems[2].Data,'\\');
            if(!filename)
              filename=DialogItems[2].Data;
            if(!(strcmp(filename,".")&&strcmp(filename,"..")))
              sprintf(DialogItems[2].Data,"%s\\",DialogItems[2].Data);
            if(!GetFullPathName(DialogItems[2].Data,sizeof(DestA),DestA,&filename))
            {
              char err3[512];
              sprintf(err3,GetMsg(mCpyErr3),DialogItems[2].Data);
              const char *MsgItems[]={"",err3,GetMsg(mOk)};
              Info.Message(Info.ModuleNumber,FMSG_ERRORTYPE|FMSG_WARNING,NULL,MsgItems,sizeof(MsgItems)/sizeof(MsgItems[0]),1);
              break;
            }
            if(DestA[strlen(DestA)-1]!='\\')
            {
              WIN32_FIND_DATAA find;
              HANDLE hFind=FindFirstFileA(DestA,&find);
              if(hFind!=INVALID_HANDLE_VALUE)
              {
                FindClose(hFind);
                if(find.dwFileAttributes&FILE_ATTRIBUTE_DIRECTORY)
                  sprintf(DestA,"%s\\",DestA);
              }
            }
            MultiByteToWideChar(CP_OEMCP,0,DestA,-1,DestW,MAX_PATH);
            if(CheckPaths(SrcW,DestW,FALSE))
              goto copyexit_err1;
            //create destination dirs
            {
              char DirA[MAX_PATH],*DirEnd;
              if(strlen(DestA)>2)
              {
                BOOL CreateFlag=FALSE; char CreateDest[MAX_PATH];
                DirEnd=DestA+3;
                do
                {
                  DirEnd=strchr(DirEnd,'\\');
                  if(DirEnd)
                  {
                    strncpy(DirA,DestA,DirEnd-DestA); DirA[DirEnd-DestA]=0;
                    DirEnd++;
                    CreateDirectoryA(DirA,NULL);
                    CreateFlag=TRUE;
                  }
                } while(DirEnd);
                if(CreateFlag)
                {
                  sprintf(CreateDest,"%s\\*",DirA);
                  WIN32_FIND_DATAA find;
                  SetLastError(0);
                  HANDLE hFind=FindFirstFileA(CreateDest,&find);
                  DWORD err=GetLastError();
                  if(hFind!=INVALID_HANDLE_VALUE)
                    FindClose(hFind);
                  if(err==ERROR_PATH_NOT_FOUND)
                  {
                    char err2[512];
                    sprintf(err2,GetMsg(mCpyErr2),DirA);
                    const char *MsgItems[]={"",err2,GetMsg(mOk)};
                    Info.Message(Info.ModuleNumber,FMSG_WARNING,NULL,MsgItems,sizeof(MsgItems)/sizeof(MsgItems[0]),1);
                    goto copyexit;
                  }
                }
              }
            }

            BOOL RealCopy=TRUE;
            send[0]=OPERATION_COPY;
            /*fill flags*/
            send[1]=0;
            if(DialogItems[4].Selected)
              send[1]|=COPYFLAG_OVERWRITE;
            if(DialogItems[5].Selected)
              send[1]|=COPYFLAG_APPEND;
            if(DialogItems[6].Selected)
              send[1]|=COPYFLAG_SKIP;
            if(DialogItems[7].Selected)
              send[1]|=COPYFLAG_REFRESH;
            if(DialogItems[8].Selected)
            {
              send[1]|=COPYFLAG_MOVE;
              //check same drive
              if((strlen(SrcA)>1)&&(strlen(DestA)>1))
                if( (toupper(SrcA[0])==toupper(DestA[0])) && (SrcA[1]==DestA[1]) && (isalpha(SrcA[0])) && (SrcA[1]==':') )
                  RealCopy=FALSE;
            }
            if(DialogItems[9].Selected)
              send[1]|=COPYFLAG_ACCESS;
            if(DialogItems[10].Selected)
              send[1]|=COPYFLAG_ROSRC;
            if(DialogItems[11].Selected)
              send[1]|=COPYFLAG_RODST;
            if(DialogItems[12].Selected)
              send[1]|=COPYFLAG_FAIL;
            if(Opt.CopyFullInfo&&RealCopy)
              send[1]|=COPYFLAG_STATISTIC;
            send[2]=PInfo.SelectedItemsNumber+2;

            #define CHECK_DEST if(CheckPaths(TestSrcW,DestW,TRUE)) goto copyexit_err1;
            #define DEST_W DestW
            #include "bcopy_find.cpp"
          }
          break;
copyexit_err1:
          {
            const char *MsgItems[]={"",GetMsg(mCpyErr1),GetMsg(mOk)};
            SetLastError(0);
            Info.Message(Info.ModuleNumber,FMSG_ERRORTYPE|FMSG_WARNING,NULL,MsgItems,sizeof(MsgItems)/sizeof(MsgItems[0]),1);
          }
copyexit:
          break;
        }
        case 2:
        {
          if(!send) break;
          if(!CheckPipeEx()) break;
          //Show dialog
          /*
            000000000011111111112222222222333333333344444444445555555555666666
            012345678901234567890123456789012345678901234567890123456789012345
          00                                                                  00
          01    Copy    01
          02    Copy "bcopy.cpp" to                                         02
          03   Ķ   03
          04    [ ] Clear R/O attribute                                     04
          05    [ ] Abort on error                                          05
          06   Ķ   06
          07    [ Delete ] [ Cancel ]                                       07
          08   ͼ   08
          09                                                                  09
            000000000011111111112222222222333333333344444444445555555555666666
            012345678901234567890123456789012345678901234567890123456789012345
          */
          static struct InitDialogItem InitDlg[]={
          /* 0*/  {DI_DOUBLEBOX,3,1,62,8,0,0,0,0,(char *)mDelDlgTitle},
          /* 1*/  {DI_TEXT,5,2,0,0,0,0,0,0,""},
          /* 2*/  {DI_TEXT,5,3,0,0,0,0,DIF_BOXCOLOR|DIF_SEPARATOR,0,""},
          /* 3*/  {DI_CHECKBOX,5,4,0,0,0,0,0,0,(char *)mDelDlgChk1},
          /* 4*/  {DI_CHECKBOX,5,5,0,0,0,0,0,0,(char *)mDelDlgChk2},
          /* 5*/  {DI_TEXT,5,6,0,0,0,0,DIF_BOXCOLOR|DIF_SEPARATOR,0,""},
          /* 6*/  {DI_BUTTON,0,7,0,0,1,0,DIF_CENTERGROUP,1,(char *)mDelDlgOk},
          /* 7*/  {DI_BUTTON,0,7,0,0,0,0,DIF_CENTERGROUP,0,(char *)mDelDlgCancel}
          };
          struct FarDialogItem DialogItems[sizeof(InitDlg)/sizeof(InitDlg[0])];
          InitDialogItems(InitDlg,DialogItems,sizeof(InitDlg)/sizeof(InitDlg[0]));
          DialogItems[3].Selected=Opt.DeleteRO;
          DialogItems[4].Selected=Opt.DeleteAbort;
          if(PInfo.SelectedItemsNumber>1)
            sprintf(DialogItems[1].Data,GetMsg(mDelDlgDeleteN+NumberType(PInfo.SelectedItemsNumber)),PInfo.SelectedItemsNumber);
          else
            sprintf(DialogItems[1].Data,GetMsg(mDelDlgDelete1),PInfo.SelectedItems[0].FindData.cFileName);
          int DlgCode=Info.DialogEx(Info.ModuleNumber,-1,-1,66,10,"Delete",DialogItems,(sizeof(InitDlg)/sizeof(InitDlg[0])),0,0,DelDialogProc,0);
          if(DlgCode==6)
          {
            send[0]=OPERATION_COPY;
            /*fill flags*/
            send[1]=COPYFLAG_DELETE|COPYFLAG_OVERWRITE;
            if(DialogItems[3].Selected)
              send[1]|=COPYFLAG_ROSRC;
            if(DialogItems[4].Selected)
              send[1]|=COPYFLAG_FAIL;
            send[2]=PInfo.SelectedItemsNumber+2;

            #define CHECK_DEST
            #define DEST_W L""
            #include "bcopy_find.cpp"
          }
          break;
        }
        case 3:
        case 4:
        {
          if((PInfo.CurrentItem >= 0) && !(PInfo.PanelItems[PInfo.CurrentItem].FindData.dwFileAttributes&FILE_ATTRIBUTE_DIRECTORY))
          {
            char SrcLA[MAX_PATH];
            char *ShortFileName=PInfo.PanelItems[PInfo.CurrentItem].FindData.cAlternateFileName;
            strcpy(SrcLA,SrcA);
            if(!ShortFileName[0])
              ShortFileName=PInfo.PanelItems[PInfo.CurrentItem].FindData.cFileName;
            strcat(SrcA,ShortFileName);
            strcat(SrcLA,PInfo.PanelItems[PInfo.CurrentItem].FindData.cFileName);
            if(MenuCode==3)
              Info.Viewer(SrcA,SrcLA,0,0,-1,-1,VF_NONMODAL);
            else
              Info.Editor(SrcA,SrcLA,0,0,-1,-1,EF_NONMODAL,1,1);
          }
          break;
        }
      }
      free(send);
    }
  }
  else
  {
    switch(MenuCode)
    {
      case 6:
      case 7:
      {
        ShowDiskMenu(MenuCode-6);
        break;
      }
      case 9:
        Configure(0);
        break;
      case 10:
      {
        DWORD size;
        SmallInfoRec *receive=NULL;
        size=0;
        {
          DWORD send[2]={OPERATION_INFO,INFOFLAG_ALL};

          DWORD dwBytesRead,dwBytesWritten;

          HANDLE hPipe=CreateFile(PIPE_NAME,GENERIC_READ|GENERIC_WRITE,0,NULL,OPEN_EXISTING,0,NULL);
          if(hPipe!=INVALID_HANDLE_VALUE)
          {
            if(WriteFile(hPipe,send,sizeof(send),&dwBytesWritten,NULL))
              if(ReadFile(hPipe,&size,sizeof(size),&dwBytesRead,NULL))
                if(size)
                {
                  receive=(SmallInfoRec *)malloc(sizeof(SmallInfoRec)*size);
                  if(receive)
                    if(!ReadFile(hPipe,receive,sizeof(SmallInfoRec)*size,&dwBytesRead,NULL))
                      size=0;
                }
            CloseHandle(hPipe);
          }
        }
        if(size&&receive)
        {
          FarMenuItem *InfoList=NULL; DWORD InfoListSize=sizeof(FarMenuItem)*size;
          InfoList=(FarMenuItem *)malloc(InfoListSize);
          if(InfoList)
          {
            int MenuCode=0;
            if(!((size==1)&&(PlgOpt.InfoMenu&INFO_MENU_ALT_1)))
            {
              char SrcA[MAX_PATH];
              int InfoMsgs[]={0,mInfoCopy,mInfoMove,mInfoDel};
              for(long unsigned int i=0;i<size;i++)
              {
                InfoList[i].Selected=InfoList[i].Checked=InfoList[i].Separator=0;
                WideCharToMultiByte(CP_OEMCP,0,receive[i].Src,-1,SrcA,MAX_PATH,NULL,NULL);
                sprintf(InfoList[i].Text,"%s%s%s",GetMsg(InfoMsgs[receive[i].type]),GetMsg(mInfoSep),FSF.TruncPathStr(SrcA,60));
              }
              InfoList[0].Selected=TRUE;
              MenuCode=Info.Menu(Info.ModuleNumber,-1,-1,0,FMENU_AUTOHIGHLIGHT|FMENU_WRAPMODE,NULL,NULL,"Contents",NULL,NULL,InfoList,size);
            }
            if(MenuCode>=0)
            {
              //Show dialog
              /*
                0000000000111111111122222222223333333333444444444455555555556666666666777777
                0123456789012345678901234567890123456789012345678901234567890123456789012345
              00                                                                            00
              01    Copy ͻ   01
              02    Copying the file                                                      02
              03    C:\Program Files\Adobe\Illustrator 9.0\Illustrator.exe                03
              04    to                                                                    04
              05    C:\TEMP\Illustrator 9.0\Illustrator.exe                               05
              06   Ķ   06
              07    Copied/Total:          19223372036854775807 / 19223372036854775807    07
              08    Start/Finish:           20.01.2001 12.01.03 /  20.01.2001 12.01.03    08
              09    Average CPS:                                  19223372036854775807    09
              10    100%     10
              11   Ķ   11
              12                    [ OK ]    [ Pause ]     [ Stop ]                      12
              13   ͼ   13
              14                                                                            14
                0000000000111111111122222222223333333333444444444455555555556666666666777777
                0123456789012345678901234567890123456789012345678901234567890123456789012345
              */
              struct FarDialogItem DialogItems[14];
              memset(DialogItems, 0, sizeof(DialogItems));
              unsigned int i=0, height=8;
              if(receive[MenuCode].type!=INFOTYPE_DELETE)
              {
                height=10;
                if(receive[MenuCode].InfoEx)
                  height=15;
              }
              //border
              DialogItems[i].Type=DI_DOUBLEBOX; DialogItems[i].X1=3; DialogItems[i].Y1=1;
              DialogItems[i].X2=72; DialogItems[i].Y2=height-2; i++;
              //Ok button
              DialogItems[i].Type=DI_BUTTON; DialogItems[i].Y1=height-3; DialogItems[i].Focus=1;
              DialogItems[i].Flags=DIF_CENTERGROUP; strcpy(DialogItems[i].Data,GetMsg(mInfoDlgOk)); i++;
              //Stop button
              DialogItems[i].Type=DI_BUTTON; DialogItems[i].Y1=height-3;
              DialogItems[i].Flags=DIF_CENTERGROUP; strcpy(DialogItems[i].Data,GetMsg(mInfoDlgStop)); i++;
              //Continue button
              DialogItems[i].Type=DI_BUTTON; DialogItems[i].Y1=height-3;
              DialogItems[i].Flags=DIF_CENTERGROUP|DIF_BTNNOCLOSE; strcpy(DialogItems[i].Data,GetMsg(mInfoDlgPause)); i++;

              DialogItems[i].Type=DI_TEXT; DialogItems[i].X1=5; DialogItems[i].Y1=2;
              switch(receive[MenuCode].type)
              {
                case INFOTYPE_COPY:
                  sprintf(DialogItems[i-4].Data,"%s",GetMsg(mCpyDlgCopyTitle));
                  sprintf(DialogItems[i].Data,"%s",GetMsg(mInfoDlgCopy));
                  break;
                case INFOTYPE_MOVE:
                  sprintf(DialogItems[i-4].Data,"%s",GetMsg(mCpyDlgMoveTitle));
                  sprintf(DialogItems[i].Data,"%s",GetMsg(mInfoDlgMove));
                  break;
                case INFOTYPE_DELETE:
                  sprintf(DialogItems[i-4].Data,"%s",GetMsg(mDelDlgTitle));
                  sprintf(DialogItems[i].Data,"%s",GetMsg(mInfoDlgDelete));
                  break;
              }
              i++;
              DialogItems[i].Type=DI_TEXT; DialogItems[i].X1=5; DialogItems[i].Y1=3; i++;
              DialogItems[i].Type=DI_TEXT; DialogItems[i].Y1=height-4; 
              DialogItems[i].Flags=DIF_BOXCOLOR|DIF_SEPARATOR; i++;

              if(receive[MenuCode].type!=INFOTYPE_DELETE)
              {
                unsigned int j=4; //i=6
                DialogItems[i].Type=DI_TEXT; DialogItems[i].X1=5; DialogItems[i].Y1=j; 
                strcpy(DialogItems[i].Data,GetMsg(mInfoDlgTo)); i++; j++;
                DialogItems[i].Type=DI_TEXT; DialogItems[i].X1=5; DialogItems[i].Y1=j; i++; j++;
                if(receive[MenuCode].InfoEx) //i=8
                {
                  DialogItems[i].Flags=DIF_BOXCOLOR|DIF_SEPARATOR;
                  for(;j<11;i++,j++)
                  {
                    DialogItems[i].Type=DI_TEXT; DialogItems[i].X1=5; DialogItems[i].Y1=j;
                  }
                }
              }
              int DlgCode=Info.DialogEx(Info.ModuleNumber,-1,-1,76,height,"Info",DialogItems,i,0,0,InfoDialogProc,receive[MenuCode].ThreadId);
              if(DlgCode==2)
              {
                DWORD send[3]={OPERATION_INFO,INFOFLAG_STOP};
                send[2]=receive[MenuCode].ThreadId;

                DWORD dwBytesWritten;

                HANDLE hPipe=CreateFile(PIPE_NAME,GENERIC_READ|GENERIC_WRITE,0,NULL,OPEN_EXISTING,0,NULL);
                if(hPipe!=INVALID_HANDLE_VALUE)
                {
                  WriteFile(hPipe,send,sizeof(send),&dwBytesWritten,NULL);
                  CloseHandle(hPipe);
                }
              }
            }
            free(InfoList);
          }
        }
        else if(PlgOpt.InfoMenu&INFO_MENU_ALT_0)
        {
          const char *MsgItems[]={"",GetMsg(mNoWork),GetMsg(mOk)};
          Info.Message(Info.ModuleNumber,0,NULL,MsgItems,sizeof(MsgItems)/sizeof(MsgItems[0]),1);
        }
        if(receive) { free(receive); receive=NULL; }
        break;
      }
    }
  }
  
  return(INVALID_HANDLE_VALUE);
}
